#include <ctype.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>


#include <stdio.h>


struct trie_node
   {
   char key;
   struct trie_node * child;
   struct trie_node * sibling;
   void * data;
   };

struct trie
   {
   struct trie_node * root;
   };

struct trie_node * trie_search_children(struct trie_node * node, char k);
void trie_visit_node(struct trie * trie, struct trie_node * node, void (*op)(char * k, void * d), char * key);
void trie_print_node(struct trie_node * node);
void trie_print_node(struct trie_node * node);
void trie_node_dump_preorder(struct trie * trie, struct trie_node * node, int depth);
void trie_node_dump_raw(struct trie_node * node, int depth);

////////////////////////////////////////////////////////////////////////////////
struct trie *
trie_new()
   {
   struct trie * trie = malloc(sizeof(struct trie));
   if (NULL == trie)
      return NULL;

   trie->root = malloc(sizeof(struct trie_node));
   if (NULL == trie->root)
      return NULL;

   trie->root->child = trie->root->sibling = NULL;
   trie->root->key = '\0';
   trie->root->data = NULL;

   return trie;
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_delete(struct trie * trie)
   {
   if (NULL == trie)
      return;

   free(trie);
   }

////////////////////////////////////////////////////////////////////////////////
char *
trie_insert(struct trie * trie, char * key, void * data)
   {
   if ( (NULL == trie) || (NULL == key) )
      return NULL;

   struct trie_node * curr = trie->root;
   struct trie_node * result;

   size_t len = strlen(key);
   for (int i = 0; i <= len; ++i)
      {
      result = trie_search_children(curr, key[i]);
      if (NULL == result)
	 {
	 // current char is not in the trie yet, allocate a new node and attach it to the children
	 struct trie_node * new = malloc(sizeof(struct trie_node));
	 if (NULL == new)
	    return NULL;
	 new->key = tolower(key[i]);
	 new->child = new->sibling = NULL;
	 new->data = NULL;

	 if (NULL == curr->child)
	    curr->child = new;
	 else
	    {
	    struct trie_node * p = curr->child;
	    while (p->sibling)
	       p = p->sibling;
	    p->sibling = new;
	    }

	 curr = new;
	 }
      else
	 curr = result;
      }

   // At this point, curr is either pointing to a new \0 node with a NULL data pointer,
   // or an existing \0 node with a valid data pointer (in which case we were asked to "insert" 
   // a key that was already present in the trie.
   // For now I am simply ignoring collisions.  We will update the data.

   curr->data = data;
   return key;
   }

////////////////////////////////////////////////////////////////////////////////
struct trie_node *
trie_search_children(struct trie_node * node, char k)
   {
   struct trie_node * p = node->child;
   if (NULL == p)
      return NULL;

   while ( (NULL != p) && (p->key != tolower(k)) )
      p = p->sibling;
   return p;
   }

////////////////////////////////////////////////////////////////////////////////
void *
trie_find(struct trie * trie, char * key)
   {
   if ( (NULL == trie) || (NULL == key) )
      return NULL;

   struct trie_node * node = trie->root;
   size_t len = strlen(key);
   for (int i = 0; i <= len; i++)
      {
      node = trie_search_children(node, key[i]);
      if (NULL == node)
	 return NULL;
      if ('\0' == key[i])
	 return node->data;
      }

   // Should never get here
   return NULL;
   }

////////////////////////////////////////////////////////////////////////////////
void *
trie_remove(struct trie * trie, char * key)
   {
   return NULL;
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_walk_keys(struct trie * trie, void (*op)(char * k, void * d))
   {
   if (NULL == trie)
      return;
   trie_visit_node(trie, trie->root, op, "");
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_visit_node(struct trie * trie, struct trie_node * node, void (*op)(char * k, void * d), char * key)
   {
   if ( (node->key == '\0') && (node != trie->root) )
      op(key, node->data);
   else
      {
      size_t len = strlen(key);
      char * k = malloc(len+2);
      if (NULL == k)
	 return;
      strcpy(k, key);
      k[len] = node->key;
      k[len+1] = '\0';
      struct trie_node * p = node->child;
      while (NULL != p)
	 {
	 trie_visit_node(trie, p, op, k);
	 p = p->sibling;
	 }
      }
   }

////////////////////////////////////////////////////////////////////////////////
void trie_dump_raw(struct trie * trie)
   { 
   if (NULL == trie)
      return;
  
   int depth = -1;
   trie_node_dump_raw(trie->root, depth);
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_node_dump_raw(struct trie_node * node, int depth)
   {
   for (int i=0; i < depth; i++)
      printf(" ");
   trie_print_node(node);
   if (node->child)
      trie_node_dump_raw(node->child, depth+1); 
   if (NULL != node->sibling)
      trie_node_dump_raw(node->sibling, depth+1);
   }

////////////////////////////////////////////////////////////////////////////////
void trie_dump(struct trie * trie)
   { 
   if (NULL == trie)
      return;
  
   int depth = -1;
   trie_node_dump_preorder(trie, trie->root, depth);
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_node_dump_preorder(struct trie * trie, struct trie_node * node, int depth)
   {
   printf("%c", node->key);
   if ( (node != trie->root) && ('\0' == node->key) )
      printf("\n");
   if (node->child)
      trie_node_dump_preorder(trie, node->child, depth+1);
   if (NULL != node->sibling)
      {
      for (int i=0; i < depth; i++)
	 printf(" ");
      trie_node_dump_preorder(trie, node->sibling, depth);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void
trie_print_node(struct trie_node * node)
   {
   printf("%p\tchild %p\tsibling %p\tkey %c\tdata %p\n", node, node->child, node->sibling, node->key, node->data);
   }
